
/*
 *  dyncnn is a deep-learning algorithm for the prediction of
 *  interacting object dynamics
 *
 *  Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 *  Written by Francois Fleuret <francois.fleuret@idiap.ch>
 *
 *  This file is part of dyncnn.
 *
 *  dyncnn is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License version 3 as
 *  published by the Free Software Foundation.
 *
 *  dyncnn is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with dyncnn.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <string.h>

#include "universe.h"

Universe::Universe(int nb_max_polygons,
                   scalar_t width, scalar_t height) : _width(width),
                                                      _height(height),
                                                      _nb_max_polygons(nb_max_polygons),
                                                      _nb_polygons(0) {
  _polygons = new Polygon *[_nb_max_polygons];
  for(int n = 0; n < _nb_max_polygons; n++) _polygons[n] = 0;
}

Universe::~Universe() {
  clear();
  delete[] _polygons;
}

void Universe::initialize_polygon(Polygon *p) {
  p->initialize(_nb_max_polygons);
}

void Universe::clear() {
  for(int n = 0; n < _nb_polygons; n++) if(_polygons[n]) delete _polygons[n];
  _nb_polygons = 0;
}

void Universe::add_polygon(Polygon *p) {
  if(_nb_polygons < _nb_max_polygons) {
    if(!p->_initialized) {
      cerr << "You can not add a non-initialized polygon." << endl;
      abort();
    }
    _polygons[_nb_polygons++] = p;
  } else {
    cerr << "Too many polygons!" << endl;
    exit(1);
  }
}

bool Universe::collide(Polygon *p) {
  for(int n = 0; n < _nb_polygons; n++)
    if(_polygons[n] && _polygons[n]->collide(p)) return true;

  return false;
}

void Universe::compute_pseudo_collisions(int nb_axis, int *nb_colliding_axis) {
  Couple couples[_nb_polygons * 2];
  int in[_nb_polygons];
  memset((void *) nb_colliding_axis, 0, _nb_polygons * _nb_polygons * sizeof(int));

  for(int a = 0; a < nb_axis; a++) {
    scalar_t alpha = M_PI * scalar_t(a) / scalar_t(nb_axis);
    scalar_t vx = cos(alpha), vy = sin(alpha);

    for(int n = 0; n < _nb_polygons; n++) {
      scalar_t *x = _polygons[n]->_x, *y = _polygons[n]->_y;
      scalar_t min = x[0] * vx + y[0] * vy, max = min;

      for(int v = 1; v < _polygons[n]->_nb_vertices; v++) {
        scalar_t s = x[v] * vx + y[v] * vy;
        if(s > max) max = s;
        if(s < min) min = s;
      }

      couples[2 * n + 0].value = min;
      couples[2 * n + 0].index = n;
      couples[2 * n + 1].value = max;
      couples[2 * n + 1].index = n;
    }

    qsort((void *) couples, 2 * _nb_polygons, sizeof(Couple), compare_couple);

    int nb_in = 0;
    memset((void *) in, 0, _nb_polygons * sizeof(int));
    for(int k = 0; k < 2 * _nb_polygons; k++) {
      int i = couples[k].index;
      in[i] = !in[i];
      if(in[i]) {
        nb_in++;
        if(nb_in > 1) {
          for(int j = 0; j < i; j++)
            if(j != i && in[j]) nb_colliding_axis[j + i * _nb_polygons]++;
          for(int j = i+1; j < _nb_polygons; j++)
            if(j != i && in[j]) nb_colliding_axis[i + j * _nb_polygons]++;
        }
      } else nb_in--;
    }
  }

  for(int i = 0; i < _nb_polygons; i++) {
    for(int j = 0; j < i; j++) {
      if(nb_colliding_axis[j + i * _nb_polygons] > nb_colliding_axis[i + i * _nb_polygons])
        nb_colliding_axis[i + i * _nb_polygons] = nb_colliding_axis[j + i * _nb_polygons];
      nb_colliding_axis[i + j * _nb_polygons] = nb_colliding_axis[j + i * _nb_polygons];
    }
  }
}

bool Universe::update(scalar_t dt) {
  bool result = false;
  apply_collision_forces(dt);
  for(int n = 0; n < _nb_polygons; n++) if(_polygons[n]) {
    _polygons[n]->apply_border_forces(dt, _width, _height);
    result |= _polygons[n]->update(dt);
  }
  return result;
}

Polygon *Universe::pick_polygon(scalar_t x, scalar_t y) {
  for(int n = 0; n < _nb_polygons; n++)
    if(_polygons[n] && _polygons[n]->contain(x, y)) return _polygons[n];
  return 0;
}

#ifdef XFIG_SUPPORT
void Universe::print_xfig(XFigTracer *tracer) {
  for(int n = 0; n < _nb_polygons; n++) {
    if(_polygons[n]) {
      _polygons[n]->color_xfig(tracer);
    }
  }
  for(int n = 0; n < _nb_polygons; n++) {
    if(_polygons[n]) {
      _polygons[n]->print_xfig(tracer);
    }
  }
}
#endif

#ifdef X11_SUPPORT
void Universe::draw(SimpleWindow *window) {
  for(int n = 0; n < _nb_polygons; n++) {
    if(_polygons[n]) {
      _polygons[n]->draw(window);
    }
  }

  for(int n = 0; n < _nb_polygons; n++) {
    if(_polygons[n]) {
      _polygons[n]->draw_contours(window);
    }
  }
}
#endif

void Universe::draw(Canvas *canvas) {
  for(int n = 0; n < _nb_polygons; n++) {
    if(_polygons[n]) {
      _polygons[n]->draw(canvas);
    }
  }

  for(int n = 0; n < _nb_polygons; n++) {
    if(_polygons[n]) {
      _polygons[n]->draw_contours(canvas);
    }
  }
}

void Universe::apply_collision_forces(scalar_t dt) {
  const int nb_axis = 2;
  int nb_collision[_nb_polygons * _nb_polygons];

  compute_pseudo_collisions(nb_axis, nb_collision);

  for(int n = 0; n < _nb_polygons; n++) if(_polygons[n])
    for(int m = 0; m < _nb_polygons; m++)
      if(m != n && _polygons[m] && nb_collision[n + _nb_polygons * m] == nb_axis)
        _polygons[n]->apply_collision_forces(dt, m, _polygons[m]);
}
